{-# LANGUAGE RankNTypes #-}
module Section2Sharing where

-- The only language extension we require here are RankNTypes,
-- for the definition of 'ClosedExpr' below.

--
-- Sharing in internal DSLs
--

-- Shallow embedding

type ExprS = Int

oneS :: ExprS
oneS = 1

(.+./) :: ExprS -> ExprS -> ExprS
(.+./) = (+)

evalS :: ExprS -> Int
evalS = id

treeS_I :: Int -> ExprS
treeS_I 0 = oneS
treeS_I n = let shared = treeS_I (n - 1) in shared .+./ shared

test_treeS_I_1 = evalS (treeS_I 2)
test_treeS_I_2 = evalS (treeS_I 30) -- fast

-- Deep embedding

data ExprD = OneD | AddD ExprD ExprD
  deriving (Show)

oneD :: ExprD
oneD   =  OneD

(.+.!) :: ExprD -> ExprD -> ExprD
(.+.!) =  AddD

evalD :: ExprD -> Int
evalD OneD         = 1
evalD (AddD e1 e2) = evalD e1 + evalD e2

textD :: ExprD -> String
textD OneD          =  "1"
textD (AddD e1 e2)  =  "(" ++ textD e1 ++ " + " ++ textD e2 ++ ")"

treeD_I :: Int -> ExprD
treeD_I 0 = oneD
treeD_I n = let shared = treeD_I (n - 1) in shared .+.! shared

test_treeD_I_1 = evalD (treeD_I 2)
test_treeD_I_2 = evalD (treeD_I 30) -- slow

-- Abstract syntax graphs

data Expr a  = One | Add (Expr a) (Expr a)
             | Var a | Let (Expr a) (a -> Expr a)

type ClosedExpr = forall a. Expr a

-- The above requires a language extension (RankNTypes, or at least
-- Rank2Types). The use of 'ClosedExpr' here is quite convenient, but
-- not absolutely essential.
--
-- If more type safety is desired for closed expressions, one can
-- wrap the polymorphic type in a newtype declaration instead:
--
-- newtype ClosedExpr = CE { unCE :: forall a. Expr a }
--
-- This leads to slightly more noisy code, though.

one   = One
(.+.) = Add

let_ :: Expr a -> (Expr a -> Expr a) -> Expr a
let_ e1 e2 =  Let e1 (\ x -> e2 (Var x))

tree_I :: Int -> ClosedExpr
tree_I 0 = one
tree_I n = let shared = tree_I (n - 1) in shared .+. shared

tree_E :: Int -> ClosedExpr
tree_E 0 = one
tree_E n = let_ (tree_E (n - 1)) (\ shared -> shared .+. shared)

-- Preserving sharing

eval :: Expr Int -> Int
eval One         = 1
eval (Add e1 e2) = eval e1 + eval e2
eval (Var n)     = n
eval (Let e1 e2) = eval (e2 (eval e1))

test_tree_I_1 = eval (tree_I 2)
test_tree_I_2 = eval (tree_I 30) -- slow
test_tree_E_1 = eval (tree_E 2)
test_tree_E_2 = eval (tree_E 30) -- fast

-- Observing sharing

text :: ClosedExpr -> String
text e = go e 0
  where
    go :: Expr String -> Int -> String
    go One          _ = "1"
    go (Add e1 e2)  c =
      "(" ++  go e1 c ++ " + " ++ go e2 c ++ ")"
    go (Var x)      _ = x
    go (Let e1 e2)  c =
      "(let " ++ v ++ " = " ++ go e1 (c + 1) ++
      " in " ++ go (e2 v) (c + 1) ++ ")"
      where v = "v" ++ show c

test_text_tree_I = text (tree_I 2)
test_text_tree_E = text (tree_E 2)

-- Inlining

inline :: Expr (Expr a) -> Expr a
inline One         = One
inline (Add e1 e2) = Add (inline e1) (inline e2)
inline (Var x)     = x
inline (Let e1 e2) = inline (e2 (inline e1))

test_text_inline_treeE = text (inline (tree_E 2))
test_eval_inline_treeE = eval (inline (tree_E 30)) -- slow

