{-# LANGUAGE RankNTypes #-}
module Section3Recursion where

import Data.Function (fix)
import Data.List (intercalate)

type ClosedExpr = forall a. Expr a

data Expr a = Lit Int | Add (Expr a) (Expr a)
            | IfZero (Expr a) (Expr a) (Expr a)
            | Var a | Let (Expr a) (a -> Expr a)
            | Lam (a -> Expr a) | App (Expr a) (Expr a)
              -- defined in 3.2 (Recursion):
            | Mu (a -> Expr a)
              -- defined in 3.3 (Mutually recursive definitions):
            | LetRec ([a] -> [Expr a]) ([a] -> Expr a)

(.+.)       =  Add
(.@.)       =  App
let_ e1 e2  =  Let e1 (\ x -> e2 (Var x))
lam_ e      =  Lam (\ x -> e (Var x))

-- Evaluation

data Value = N Int | F (Value -> Value)

instance Show Value where
  show (N n) = "N " ++ show n
  show (F f) = "F <function>"

eval :: Expr Value -> Value
eval (Lit i)           = N i
eval (Add e1 e2)       = add (eval e1) (eval e2)
eval (IfZero e1 e2 e3) = ifZero (eval e1) (eval e2) (eval e3)
eval (Var x)           = x
eval (Let e1 e2)       = eval (e2 (eval e1))
eval (Lam e)           = F (\ v -> eval (e v))
eval (App e1 e2)       = app (eval e1) (eval e2)
-- defined in 3.2 (Recursion):
eval (Mu e)            = fix (\ v -> eval (e v))
-- defined in 3.3 (Mutually recursive definitions):
eval (LetRec es e)     = let rs = map eval (es rs) in eval (e rs)

add (N m) (N n)    = N (m + n)
ifZero (N n) v1 v2 = if n == 0 then v1 else v2
app (F f) v        = f v

example :: ClosedExpr
example =
  let_  (lam_ (\ x -> x .+. Lit (-1))) (\ dec ->
  let_  (lam_ (\ f -> lam_ (\ x -> f .@. (f .@. x)))) (\ twice ->
        (twice .@. twice .@. dec .@. Lit 10)))

exampleHaskell :: Int
exampleHaskell =
  let dec     x = x - 1
      twice f x = f (f x)
  in  twice twice dec 10

test_eval_example :: Value
test_eval_example = eval example

-- Recursion

-- Mu and eval for Mu already defined above

mu_ e = Mu (\ x -> e (Var x))

mul :: ClosedExpr
mul =  lam_ (\ m -> mu_ (\ rec -> lam_ (\ n ->
            IfZero  n  (Lit 0) (m .+. (rec .@. (n .+. Lit (-1)))))))

test_mul = eval (mul .@. Lit 3 .@. Lit 7)

letrec :: (Expr a -> Expr a) -> (Expr a -> Expr a) -> Expr a
letrec e1 e2 = Let (Mu (\ x -> e1 (Var x))) (\ x -> e2 (Var x))

-- Mutually recursive definitions

evenOddHaskell :: Int
evenOddHaskell =
  let dec  x     = x - 1
      even x t e = if x == 0 then t else odd  (dec x) t e
      odd  x t e = if x == 0 then e else even (dec x) t e
  in even 4 1 0

-- LetRec and eval for LetRec already defined above

letrec_ ::  ([Expr a] -> [Expr a]) ->
            ([Expr a] -> Expr a) -> Expr a
letrec_ es e = LetRec (\ xs -> es (map Var xs))
                      (\ xs -> e  (map Var xs))

evenOdd :: ClosedExpr
evenOdd = letrec_ (\ ~[dec, even, odd] ->
             [ lam_ (\ x -> x .+. Lit (-1))
             , lam_ (\ x -> lam_ (\ t -> lam_ (\ e ->
                 IfZero x t  (odd   .@. (dec .@. x) .@. t .@. e))))
             , lam_ (\ x -> lam_ (\ t -> lam_ (\ e ->
                 IfZero x e  (even  .@. (dec .@. x) .@. t .@. e))))
             ])
             (\ [dec, even, odd] -> even .@. Lit 4 .@. Lit 1 .@. Lit 0)

test_eval_evenOdd = eval evenOdd

-- Not shown in paper:

text :: ClosedExpr -> String
text e = go e 0
  where
    go :: Expr String -> Int -> String
    go (Lit n)           _ = show n
    go (Add e1 e2)       c =
      "(" ++  go e1 c ++ " + " ++ go e2 c ++ ")"
    go (IfZero e1 e2 e3) c =
      "(ifzero " ++ go e1 c ++ " then " ++ go e2 c ++
                               " else " ++ go e3 c ++ ")"
    go (Var x)           _ = x
    go (Let e1 e2)       c =
      "(let " ++ v ++ " = " ++ go e1 (c + 1) ++
      " in " ++ go (e2 v) (c + 1) ++ ")"
      where v = "v" ++ show c
    go (Lam e)           c =
      "(\\ " ++ v ++ " -> " ++ go (e v) (c + 1) ++ ")"
      where v = "v" ++ show c
    go (App e1 e2)       c =
      "(" ++ go e1 c ++ " " ++ go e2 c ++ ")"
    go (Mu e)            c =
      "(mu " ++ v ++ " -> " ++ go (e v) (c + 1) ++ ")"
      where v = "v" ++ show c
    go (LetRec es e)     c =
      "(let { " ++ intercalate "; " ds ++
      " } in " ++ go (e vs) c' ++ ")"
      where
        n   = length (es undefined) -- look ahead at length of result
        vs  = take n (map (\ i -> "v" ++ show i) (enumFrom c))
        c'  = c + n
        ds  = zipWith  (\v e -> v ++ " = " ++ go e c') vs (es vs)

test_text_mul     = text mul
test_text_evenOdd = text evenOdd

-- Reusing native let syntax
--
-- Because the corresponding code depends on the data-reify package,
-- it is contained in a separate module called Section34NativeLet.

