{-# LANGUAGE RankNTypes, KindSignatures, GADTs, ScopedTypeVariables #-}
module Section5TypedASGs where

import Data.Function (fix)
import Data.List (intercalate)
import Section4TypedLists

-- Well-typed Abstract Syntax Graphs

type ClosedExpr t = forall f. Expr f t

data Expr (f :: * -> *) :: * -> *  where
    Lit    :: Int -> Expr f Int
    Add    :: Expr f Int -> Expr f Int -> Expr f Int
    IfZero :: Expr f Int -> Expr f t -> Expr f t -> Expr f t
    Var    :: f t -> Expr f t
    Lam    :: (f t1 -> Expr f t2) -> Expr f (t1 -> t2)
    App    :: Expr f (t1 -> t2) -> Expr f t1 -> Expr f t2
    LetRec :: CList ts => (TList f ts -> TList (Expr f) ts) ->
                          (TList f ts -> Expr f t) -> Expr f t

-- Well-typed evaluator

eval :: Expr I t -> t
eval (Lit i)           =  i
eval (Add e1 e2)       =  eval e1 + eval e2
eval (IfZero e1 e2 e3) =  if eval e1 == 0 then eval e2 else eval e3
eval (Var x)           =  unI x
eval (Lam e)           =  eval . e . I
eval (App e1 e2)       =  (eval e1) (eval e2)
eval (LetRec es e)     =  eval (e (fix (tmap (I . eval) . es)))

(.+.) = Add
(.@.) = App  -- not shown in paper
one   = Lit 1

lam_ :: (Expr f t1 -> Expr f t2) -> Expr f (t1 -> t2)
lam_ e = Lam (\ x -> e (Var x))

letrec_ :: CList ts => (TList (Expr f) ts -> TList (Expr f) ts) ->
                       (TList (Expr f) ts -> Expr f t) -> Expr f t
letrec_ es e = LetRec  (\ xs -> es (tmap Var xs))
                       (\ xs -> e  (tmap Var xs))

let_ :: Expr f t1 -> (Expr f t1 -> Expr f t2) -> Expr f t2
let_ e1 e2 = letrec_ (\ _ -> e1 ::: TNil) (\ (x ::: TNil) -> e2 x)

-- Not shown in paper:
mu_ :: forall f t. (Expr f t -> Expr f t) -> Expr f t
mu_ e = letrec_ decl (\ (x ::: TNil) -> x)
  where
    decl :: TList (Expr f) (t, ()) -> TList (Expr f) (t, ())
    decl (x ::: TNil) = e x ::: TNil

tree_I :: Int -> ClosedExpr Int
tree_I 0 = one
tree_I n = let shared = tree_I (n - 1) in shared .+. shared

tree_E :: Int -> ClosedExpr Int
tree_E 0 = one
tree_E n = let_ (tree_E (n - 1)) (\ shared -> shared .+. shared)

test_tree_I_1 = eval (tree_I 2)
test_tree_I_2 = eval (tree_I 30) -- slow
test_tree_E_1 = eval (tree_E 2)
test_tree_E_2 = eval (tree_E 30) -- fast

-- Fails due to polymorphic use of twice:
-- example :: ClosedExpr Int
-- example =
--   let_  (lam_ (\ x -> x .+. Lit (-1))) (\ dec ->
--   let_  (lam_ (\ f -> lam_ (\ x -> f .@. (f .@. x)))) (\ twice ->
--         (twice .@. twice .@. dec .@. Lit 10)))

-- This version works:
example' :: ClosedExpr Int
example' =
  let_  (lam_ (\ x -> x .+. Lit (-1))) (\ dec ->
  let_  (lam_ (\ f -> lam_ (\ x -> f .@. (f .@. x)))) (\ twice ->
        (twice .@. (twice .@. dec) .@. Lit 10)))

test_eval_example' = eval (example')

-- Printing terms

text :: ClosedExpr t -> String
text e = go e 0
  where
    go :: Expr (K String) t -> Int -> String
    go (Var x)        _  =  unK x
    go (Lam e)        c  =
      "(\\ " ++ v ++ " -> " ++ go (e (K v)) (c + 1)
      where v = "v" ++ show c
    go (LetRec es e)  c  =
      "(let { " ++ intercalate "; " ds ++
      " } in " ++ go (e vs) c' ++ ")"
      where
        vs  =  tmapK (\i -> "v" ++ show i) (tenumFrom c)
        c'  =  c + tlength vs
        ds  =  ttoList $
               tzipWith  (\ (K v) e -> K (v ++ " = " ++ go e c'))
                         vs (es vs)
    -- Cases not shown in paper:
    go (Lit n)        _  =  show n
    go (Add e1 e2)    c  =
      "(" ++ go e1 c ++ " + " ++ go e2 c ++ ")"
    go (IfZero e1 e2 e3) c =
      "(if " ++ go e1 c ++ " == 0 then " ++ go e2 c ++ " else " ++ go e3 c ++ ")"
    go (App e1 e2)    c  =
      "(" ++ go e1 c ++ " " ++ go e2 c ++ ")"

ttoList :: TList (K a) ts -> [a]
ttoList TNil         = []
ttoList (K x ::: xs) = x : ttoList xs

test_text_example' = text example'

-- Not shown in paper: translating evenOdd
--
-- This unfortunately requires a bit more type information
-- that we would desire.

type LetRec f ts t = (TList (Expr f) ts -> TList (Expr f) ts) ->
                     (TList (Expr f) ts -> Expr f t) -> Expr f t

evenOdd :: ClosedExpr Int
evenOdd =  (letrec__ :: LetRec f  (Int -> Int,
                                  (Int -> Int -> Int -> Int,
                                  (Int -> Int -> Int -> Int,
                                  ()))) Int)
           (\ (dec ::: even ::: odd ::: TNil) ->
                    lam_ (\ x -> x .+. Lit (-1))
               :::  lam_ (\ x -> lam_ (\ t -> lam_ (\ e ->
                      IfZero x t  (odd   .@. (dec .@. x) .@. t .@. e))))
               :::  lam_ (\ x -> lam_ (\ t -> lam_ (\ e ->
                      IfZero x e  (even  .@. (dec .@. x) .@. t .@. e))))
               :::  TNil)
           (\ (dec ::: even ::: odd ::: TNil) ->
               even .@. Lit 4 .@. Lit 1 .@. Lit 0)

-- Note that we do not require a lazy pattern match. This is
-- achieved by 'letrec__', which is like 'letrec_', but calls
-- 'tlazy':
letrec__ :: CList ts => LetRec f ts t
letrec__ es e = LetRec (\ xs -> es (tlazy (tmap Var xs)))
                       (\ xs -> e         (tmap Var xs))

-- The function 'tlazy' makes a list lazy in its spine. This works only
-- because we have a type-level list and therefore the spine structure
-- is actually statically known. We can therefore match on the statically
-- known structural representation rather than on the dynamic list:
tlazy :: CList ts => TList f ts -> TList f ts
tlazy = go cList
  where
    go :: RList ts -> TList f ts -> TList f ts
    go RNil        _  = TNil
    go (RCons rs)  xs = thead xs ::: go rs (ttail xs)

test_eval_evenOdd = eval evenOdd
test_text_evenOdd = text evenOdd
